/*
	< DanaSah is a chess engine winboard. >
    Copyright (C) <2008>  <Pedro Castro Elgarresta>  
	E-mail <pecastro@msn.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Funciones para acceder a las bitbases. Gracias a Daniel Shawul.
Para ms informacin sobre esta funciones deberamos recurrir a Scorpio.
Solo lo he hecho funcionar para windows, en linux tendriamos que llamar
en lugar de a egbbdll.dll a otro archivo.
*/

#include <stdio.h>
#include <string.h>
#include <limits.h>

#if (defined _WIN32) || defined (_WIN64)
	#include <windows.h>
	#define DIR_SEP '\\'
	#define EGBB_NAME_GENERIC "egbbdll.dll"
	//funny isn't it? :
	//gets even funnier,as gnu-c on windows already has PATH_MAX
	#ifndef PATH_MAX
		#define PATH_MAX MAX_PATH
	#endif
	#ifdef _WIN32
		#define EGBB_NAME "egbbdll32.dll"
	#else
	#ifdef _WIN64
		#define EGBB_NAME "egbbdll64.dll"
	#endif
#endif
#else
	#define DIR_SEP '/'
	#include <dlfcn.h>
	#define EGBB_NAME "egbbso.so"
	#define HMODULE void*
	#define LoadLibrary(x) dlopen(x,RTLD_LAZY)
	#define FreeLibrary(x) dlclose(x)
	#define GetProcAddress dlsym
#endif
#include "egbb.h"
typedef void (*PLOAD_EGBB) (char* path,int cache_size,int load_options);

/*
Load the dll and get the address of the load and probe functions.
*/


int LoadEgbbLibrary(char* main_path, int egbb_cache_size,PPROBE_EGBB *probe_it) {
	static HMODULE hmod;
	PLOAD_EGBB load_egbb;
	int len;
	char path[PATH_MAX];
	char dll_name[PATH_MAX];
	
	strncpy(path,main_path,sizeof(path)-1);
	path[sizeof(path)-1]='\0';
	//add \\ or / to path,if needed...
	len=(int)strlen(path)-1;
	if(len<0)return 0;
	if(path[len]!=DIR_SEP){ path[len+1]=DIR_SEP; path[len+2]='\0';}

	strncpy(dll_name,path,sizeof(dll_name));
	strcat(dll_name,EGBB_NAME);
//printf("dll %s\n path %s\n",path,dll_name);
	if (hmod)
		FreeLibrary(hmod);
#ifdef EGBB_NAME_GENERIC
	if(!(hmod = LoadLibrary(dll_name))) {
		//try again with first option
		strncpy(dll_name,path,sizeof(dll_name));
		strcat(dll_name,EGBB_NAME_GENERIC);
	}
#endif
if(hmod = LoadLibrary(dll_name)) {
		load_egbb = (PLOAD_EGBB) GetProcAddress(hmod,"load_egbb_5men");
     	*probe_it = (PPROBE_EGBB) GetProcAddress(hmod,"probe_egbb_5men");
		if(load_egbb==0)return 0;
		if(probe_it==0) return 0;
        load_egbb(path, egbb_cache_size*1024*1024,LOAD_4MEN);
		return 1;
	} else {
		printf("EgbbProbe not Loaded!\n");
		return 0;
	}
}
