/**
*Buzz Chess Engine
*moveordering.c
*
*Orders move to help alphabeta searching.
*
*Copyright (C) 2008 Pradu Kannan.
**/
#include "xboard.h"

#include "search.h"
#include "defs.h"
#include "moveordering.h"
#include "bitinstructions.h"
#include "hash.h"
#include "eval.h"
#include <stdlib.h>

//needed
#define CHECK_FLAG

move killer[MAX_SEARCH_DEPTH+1];
#ifdef TWO_KILLERS
	move killer2[MAX_SEARCH_DEPTH+1];
#endif
#ifdef USE_NULL_KILLER
	move null_killer[MAX_SEARCH_DEPTH+1];
#endif

//[piece][side][to]
int history[NUMPIECETYPES][2][64];
int hvisits[NUMPIECETYPES][2][64];

/**************************
*Private Functions and Data
**************************/

/*#ifdef max
	#undef max
#endif
INLINE int max(int a, int b)
{
	if(a>b) return a;
	return b;
}

#ifdef min
	#undef min
#endif
INLINE int min(int a, int b)
{
	if(a<b) return a;
	return b;
}*/

void orderMoves(const board* pos, moveList* ml, int dfr,move hashMove)
{
	unsigned int i;
	move t_move;
	U64 passedPawns;
 
#ifdef CHECK_FLAG
	//possible attack squares to the opponent king
	U64 KAttack[NUMPIECETYPES+1];
	KAttack[B] = Bmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[R] = Rmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[N] = Nmoves(pos->KingPos[pos->xside]);
	KAttack[P] = Pcaps(pos->KingPos[pos->xside],pos->xside);
	KAttack[Q]=  KAttack [B]|KAttack[R];
	KAttack[K]= U64EMPTY;
	KAttack[E]= U64EMPTY;
#endif
	//don't probe for hashmove,the caller should 
	//if(hashMove==NULLMOVE)
	//	if(probeHash(pos->hashkey,&ownprobe))
	//		hashMove=extractMove(ownprobe.data);

	//Generate Passers
	{
		U64 Pawns[2]={piecesWHITE(*pos,P),piecesBLACK(*pos,P)};
		
		if(pos->side==WHITE)
		{
			U64 blockers = Pawns[BLACK]|
				((Pawns[BLACK]&C64(0x7F7F7F7F7F7F7F7F))>>7)|
				((Pawns[BLACK]&C64(0xFEFEFEFEFEFEFEFE))>>9);
			U64 doubled = fillDown(Pawns[WHITE]);
			passedPawns = Pawns[WHITE] & ~(doubled | fillDown2(blockers));
		}
		else
		{
			U64 blockers = Pawns[WHITE]|
				((Pawns[WHITE]&C64(0x7F7F7F7F7F7F7F7F))<<9)|
				((Pawns[WHITE]&C64(0xFEFEFEFEFEFEFEFE))<<7);
			U64 doubled = fillUp(Pawns[BLACK]);
			passedPawns = Pawns[BLACK] & ~(doubled | fillUp2(blockers));
		}
	}

	for(i=0;i<ml->moveCount;i++)
	{
		//Extra flags

		ml->moves[i].score = 0;
		t_move=ml->moves[i].m;

		if(toBit(extractFrom(t_move)) & passedPawns) ml->moves[i].m|=encodeExtra(MO_PASSER);
		//if(toBit(extractTo(t_move)) & passedPawns) ml->moves[i].m|=encodeExtra(MO_PASSER);//logical,but test seemed worse
#ifdef CHECK_FLAG
		//checks,non-switch code:
		if(toBit(extractTo(t_move)) & KAttack[extractPiece(t_move)]){
			ml->moves[i].m|=encodeExtra(MO_CHECK);
		}
#endif

		//Hash Move

		if(t_move==hashMove)
		{	
			//if(t_move!=NULLMOVE)
			assert(t_move!=NULLMOVE);
				ml->moves[i].score=MOVEORDERING_HASHMOVE;
			continue;
		}

		//captures and promotions! (neither should be stored in the killerlists)
		if((extractCapture(t_move)!=E) || (extractPromotion(t_move)!=E))
		{
			ml->moves[i].score=SEE(pos,t_move);
			if(ml->moves[i].score>=0){ //winning cap
				ml->moves[i].score+=MOVEORDERING_CAPTURE;
			}
			else {
				ml->moves[i].score-=MOVEORDERING_CAPTURE;
			}
		continue;
		}

		//Killers
		if(t_move==killer[dfr])
		{
			ml->moves[i].score=MOVEORDERING_KILLER;
			continue;
		}

#ifdef TWO_KILLERS
		if(t_move==killer2[dfr])
		{
			ml->moves[i].score=MOVEORDERING_KILLER2;
			continue;
		}
#endif

#if 0
		if(dfr>=2 && t_move==killer[dfr-2] && SEE(pos,t_move)>=0)
		{
			ml->moves[i].score=(MOVEORDERING_KILLER2-3);
			continue;
		}
		else if(dfr>=2 && t_move==killer2[dfr-2] && SEE(pos,t_move)>=0)
		{
			ml->moves[i].score=(MOVEORDERING_KILLER2-4);
			continue;
		}
#endif
#ifdef USE_NULL_KILLER
		if(t_move==null_killer[dfr])
		{
			ml->moves[i].score=MOVEORDERING_NULLKILLER;
			continue;
		}
#endif

		if((t_move==KingsideCastle) || (t_move==QueensideCastle))
			{
			ml->moves[i].score=MOVEORDERING_CASTLING;
			}
#if 0
		//Passed Pawn Pushes
		if(toBit(extractFrom(t_move)) & passedPawns)
		{
			ml->moves[i].score=MOVEORDERING_PASSER;
			continue;
		}
#endif


#if 0
		//Mobility Addition
		switch(extractPiece(ml->moves[i].m))
		{
		case N:
		ml->moves[i].score+=(popcnt(Nmoves(extractTo(ml->moves[i].m)))
							-popcnt(Nmoves(extractFrom(ml->moves[i].m))));
			break;
		case B:
			ml->moves[i].score+=(popcnt(Bmoves(extractTo(ml->moves[i].m),pos->AllPieces))
								-popcnt(Bmoves(extractFrom(ml->moves[i].m),pos->AllPieces)));
			break;
//		case R:
//			ml->moves[i].score+=(popcnt(Rmoves(extractTo(ml->moves[i].m),pos->AllPieces))
//								-popcnt(Rmoves(extractFrom(ml->moves[i].m),pos->AllPieces)));
//			break;
//		case Q:
//			ml->moves[i].score+=(popcnt(Qmoves(extractTo(ml->moves[i].m),pos->AllPieces))
//								-popcnt(Qmoves(extractFrom(ml->moves[i].m),pos->AllPieces)));
//			break;
		default:
			break;
		}
#endif
		//History Heuristic
		ml->moves[i].score+=history[extractPiece(t_move)][pos->side][extractTo(t_move)];
		/*if(dfr<=4)
			ml->moves[i].score = -qsearch(*pos, ml->moves[i].m, -INF, INF, dfr);
		else ml->moves[i].score+=history[extractPiece(ml->moves[i].m)][pos->side][extractTo(ml->moves[i].m)];
		*/
	}

}
#if 0
//same as orderQMoves,but sets MO_CHECK flag if the move checks,its pretty cheap
void orderQRMoves(const board* pos, moveList* ml)
{
	unsigned int i;
	//possible attack squares to the opponent king
	U64 KAttack[NUMPIECETYPES+1];
	KAttack[B] = Bmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[R] = Rmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[N] = Nmoves(pos->KingPos[pos->xside]);
	KAttack[P] = Pcaps(pos->KingPos[pos->xside],pos->xside);
	KAttack[Q]=  KAttack [B]|KAttack[R];
	KAttack[K]= U64EMPTY;
	KAttack[E]= U64EMPTY;
	for(i=0;i<ml->moveCount;i++){
		ml->moves[i].score=SEE(pos,ml->moves[i].m);
		if(toBit(extractTo(ml->moves[i].m)) & KAttack[extractPiece(ml->moves[i].m)])
			ml->moves[i].m|=encodeExtra(MO_CHECK);
	}
	sortMoveList(ml);
}
#endif

//order captures+promotions(+checks)
void orderQMoves(const board* pos, moveList* ml)
{
	unsigned int i;

	for(i=0;i<ml->moveCount;i++){
		ml->moves[i].score=SEE(pos,ml->moves[i].m);
//		store positive captures/promos above positive checks
		if(ml->moves[i].score>=0){ //winning material? or not giving it away by moving a piece to a bad location(!)
		if((extractCapture(ml->moves[i].m)!=E) || (extractPromotion(ml->moves[i].m)!=E))
				ml->moves[i].score+=MOVEORDERING_CAPTURE;
			//else	ml->moves[i].score=100; //checking moves
		}
	}
	sortMoveList(ml);
}
//same but with checking moves
void orderQCMovesnu(const board* pos, moveList* ml)
{
	unsigned int i;
	U64 KAttack[NUMPIECETYPES+1];
	KAttack[B] = Bmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[R] = Rmoves(pos->KingPos[pos->xside],pos->AllPieces);
	KAttack[N] = Nmoves(pos->KingPos[pos->xside]);
	KAttack[P] = Pcaps(pos->KingPos[pos->xside],pos->xside);
	KAttack[Q]=  KAttack [B]|KAttack[R];
	KAttack[K]= U64EMPTY;
	KAttack[E]= U64EMPTY;

	for(i=0;i<ml->moveCount;i++){

		ml->moves[i].score=SEE(pos,ml->moves[i].m);
		if(toBit(extractTo(ml->moves[i].m)) & KAttack[extractPiece(ml->moves[i].m)]){
			ml->moves[i].m|=encodeExtra(MO_CHECK);
			ml->moves[i].score+=50;
		}
//		store positive captures/promos above positive checks
		if(ml->moves[i].score>=0){ //winning material? or not giving it away by moving a piece to a bad location(!)
		if((extractCapture(ml->moves[i].m)!=E) || (extractPromotion(ml->moves[i].m)!=E))
				ml->moves[i].score+=MOVEORDERING_CAPTURE;
		}
	}
	sortMoveList(ml);
}
move nextmove(moveList *ml,int *score)
{
	unsigned int n;
	unsigned int bestpick=1000;//not possible,i guess

	int  best_score=-(MOVEORDERING_HASHMOVE-1);
	//weed out this one first:
	if(ml->moveCount==1) {ml->moveCount=0;*score=ml->moves[0].score;return ml->moves[0].m;}
	//walk move list and get move with highest score
	for(n=0;n<ml->moveCount;n++){
		if((best_score<=ml->moves[n].score)){/*< or <= */
			best_score=ml->moves[n].score;
			bestpick=n;			
		}
	}
	if(bestpick>=1000) return NULLMOVE;
//	if(bestpick<1000){
		*score=best_score; // return score
		ml->moves[bestpick].score=-(MOVEORDERING_HASHMOVE);//don't pick again
		return ml->moves[bestpick].m;
//	}
//	return NULLMOVE;
}

void sortMoveList(moveList* ml)
{
	int i;
	for(i=(ml->moveCount-1);i>0;i--)
	{
		int j;
		for(j=1;j<=i;j++)
			if(ml->moves[j-1].score < ml->moves[j].score)
			{
				smove temp = ml->moves[j-1];
				ml->moves[j-1] = ml->moves[j];
				ml->moves[j] = temp;
			}
	}
}

#ifdef USE_NULL_KILLER
void storeNullKiller(const move m, const int dfr)
{
	if(extractCapture(m)==E && (extractPromotion(m)==E) )
		null_killer[dfr]=m;
}
#endif

void clearHistory()
{
	int i,j,k;
	for(i=0;i<NUMPIECETYPES;i++)
		for(j=0;j<2;j++)
			for(k=0;k<64;k++){
				history[i][j][k]=0;
				hvisits[i][j][k]=0;
			}
}

//Thanks to Andrew Fan for this hint
void clearKillers()
{
	int i;
	for(i=0;i<MAX_SEARCH_DEPTH+1;i++)
	{
		killer[i]=NULLMOVE;
#ifdef USE_NULL_KILLER
		null_killer[i]=NULLMOVE;
#endif
#ifdef TWO_KILLERS
		killer2[i]=NULLMOVE;
#endif
	}
}
void storeGoodMove(const move m,const int depth,const int side,const int dfr)
{
	if(extractCapture(m)==E && (extractPromotion(m)==E))
	{
		//killer part:
#ifndef TWO_KILLERS
#ifdef USE_NULL_KILLER
		if(null_killer[dfr]!=m)
#endif
		killer[dfr]=m;
#else
#ifdef USE_NULL_KILLER
		//it could be equal to the second,ignore for now
		if(null_killer[dfr]!=m)
#endif
		if(killer[dfr]!=m)
		{
			killer2[dfr]=killer[dfr];
			killer[dfr]=m;
		}
#endif
		//history part
		history[extractPiece(m)][side][extractTo(m)]+=depth*depth;
		if(history[extractPiece(m)][side][extractTo(m)]>=MOVEORDERING_MAXHIST){
			int i,j,k;
			for(i=0;i<NUMPIECETYPES;i++)
				for(j=0;j<2;j++)
					for(k=0;k<64;k++)
						history[i][j][k]/=16384;

		}
	}
}

void updatecount(move m,int side){
	if(extractCapture(m)==E && (extractPromotion(m)==E))
	{
	hvisits[extractPiece(m)][side][extractTo(m)]=1;
	}
}

bool was_searched(move m,int side){
	if(extractCapture(m)==E && (extractPromotion(m)==E))
	{
		if(hvisits[extractPiece(m)][side][extractTo(m)]<=0) return false;
	}
	return true;
}

bool looks_lousy(move m,int depth,int side)
{
		if(history[extractPiece(m)][side][extractTo(m)]<(depth+depth)) return true;
		return false;
}
void storeKiller(const move m, const int dfr)
{
	if(extractCapture(m)==E && (extractPromotion(m)==E) )
	{
		#ifndef TWO_KILLERS
			killer[dfr]=m;
		#else
			if(killer[dfr]!=m)
			{
				killer2[dfr]=killer[dfr];
				killer[dfr]=m;
			}
		#endif
	}
}

void AdjustMoveScore(const move m, const int depth, const int side)
{
		history[extractPiece(m)][side][extractTo(m)]--;//or -depth or -depth*depth 
		if(history[extractPiece(m)][side][extractTo(m)]<=-MOVEORDERING_CAPTURE){
			int i,j,k;
			for(i=0;i<NUMPIECETYPES;i++)
				for(j=0;j<2;j++)
					for(k=0;k<64;k++)
						history[i][j][k]/=16384;
		}
}

//full static exchange evaluvation
int SEE(const board* pos, const move m)
{
	//The variables are set to the inital state where the move in question has been played
	int SEEscore=pieceValue[extractCapture(m)]; //The score to be returned
	int SEEmax, SEEmin; //Maximum and minimum SEE scores (alphbeta approach)
	int target=extractTo(m); //The square that we are doing a SEE on
	int currentPieceValue=pieceValue[extractPiece(m)]; //Value of the piece currently on the target square
	U64 used=toBit(extractFrom(m)); //Used attackers
	//Battery attacks will be calculated instead of regular attacks
	//Piece order isn't a problem (except when dealing with Queens)
	U64 occupancy=(pos->AllPieces^used) &
		 ~(
			(pos->Pieces[B]&BmovesNoOcc(target))|(pos->Pieces[R]&RmovesNoOcc(target))|
			( (piecesBLACK(*pos,P)&Pcaps(target,WHITE)) |  (piecesWHITE(*pos,P)&Pcaps(target,BLACK)) )
		  ); 

	U64 attackers; //attackers of each piece type from both sides
	U64 attackersSide; //attackers for each piece type for a particular side
	U64 attackersPiece; //attackers for a particular color and a particular type

	//handle enpassant and promotion
	if(currentPieceValue==Pval)
	{
		if(extractEP(m))
		{
			used|=toBit(extractEP(m));
			occupancy^=toBit(extractEP(m));
			SEEscore=Pval;
		}
		else if(extractPromotion(m)!=E)
		{
			currentPieceValue=pieceValue[extractPromotion(m)];
			SEEscore+=currentPieceValue-Pval;
		}
	}

	//these are the bounds, we will be doing an alphabeta-like search in SEE
	SEEmax=SEEscore; //upperbound
	SEEmin=-INF; //lowerbound

	//Generate attackers
	attackers=attacksToOcc(*pos,target,occupancy)&~used;
	//Loop Through Opponent Captures
	while(attackersSide=attackers&pos->PiecesSide[pos->xside])
	{
		assert(!(attackers&used));
		SEEscore-=currentPieceValue;
		if(SEEscore>=SEEmax) return SEEmax;
		if(SEEscore>SEEmin) SEEmin=SEEscore;

		if(attackersPiece=attackersSide&pos->Pieces[P]) //Pawn
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece); 
			currentPieceValue=Pval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[N]) //Knight
		{
			attackers^=LSB(attackersPiece);
			currentPieceValue=Nval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[B]) //Bishop
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece);
			currentPieceValue=Bval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[R]) //Rook
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece);
			currentPieceValue=Rval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[Q]) //Queen
		{
			used|=LSB(attackersPiece);
			occupancy^=LSB(attackersPiece);
			attackers=attacksBQRToOcc(*pos,target,occupancy);
			attackers&=~used;
			currentPieceValue=Qval;
		}
		else //King
		{
			attackers^=toBit(pos->KingPos[pos->xside]);
			currentPieceValue=Kval;
		}
		//Loop Through My Captures
		if(!(attackersSide=attackers&pos->PiecesSide[pos->side])) break;
		assert(!(attackers&used));

		SEEscore+=currentPieceValue;
		if(SEEscore<SEEmin) return SEEmin;
		if(SEEscore<SEEmax) SEEmax=SEEscore;

		if(attackersPiece=attackersSide&pos->Pieces[P]) //Pawn
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece); 
			currentPieceValue=Pval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[N]) //Knight
		{
			attackers^=LSB(attackersPiece);
			currentPieceValue=Nval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[B]) //Bishop
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece);
			currentPieceValue=Bval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[R]) //Rook
		{
			attackers^=LSB(attackersPiece);
			used|=LSB(attackersPiece);
			currentPieceValue=Rval;
		}
		else if(attackersPiece=attackersSide&pos->Pieces[Q]) //Queen
		{
			used|=LSB(attackersPiece);
			occupancy^=LSB(attackersPiece);
			attackers|=attacksBQRToOcc(*pos,target,occupancy);
			attackers&=~used;
			currentPieceValue=Qval;
		}
		else //King
		{
			attackers^=toBit(pos->KingPos[pos->side]);
			currentPieceValue=Kval;
		}
	}
	if(SEEscore<SEEmin) return SEEmin;
	if(SEEscore>SEEmax) return SEEmax;
	return SEEscore;
}

void clearMoveOrderingData()
{
	clearHistory();
	clearKillers();
	resetHashTable();
}
