/**
*Buzz Chess Engine
*nalimov.cpp
*
*Probes Nalimov endgame tablebases. The code is hacked up beyond repair.
*
*Copyright (C) 2008 Pradu Kannan. All rights reserved.
**/

/*
*Dependancies
*/
#include "defs.h"
#include "board.h"
#include "eval.h"
#include "bitinstructions.h"
#include <stdlib.h>
#include <string.h>

/*
*Data
*/
#if defined (__cplusplus)  //(_MSC_VER)
extern "C" {
#endif
int NalimovNumPieces = 0;
void* NalimovCache = NULL;
#if defined (__cplusplus)  //(_MSC_VER)
}
#endif


/*
*Stuff
*/

#define NALIMOV_MATE (L_bev_mi1+L_bev_mimin)

//INDEX must be unsigned int for 5-men probing and it
//must be U64 for 6-men probing
typedef unsigned int INDEX;

//The fastest integer type on the processor
#define square int

#define NEW   //Use new tablebase format
#define XX 64 //an invalid square

#define T41_INCLUDE //Probe 4+1 tablebases
//#define DEBUG       //(hack) NDEBUG is not automatically defined

//For Nalimov code, white is 0, black is 1, and P=1,N=2,B=3,R=4,Q=5,K=6
//A1=0, B1=1 ... and H8=63

#define  C_PIECES  3    /* Maximum # of pieces of one color OTB */
#define  SqFindKing(psq)          ((psq)[C_PIECES*(5)])
#define  SqFindFirst(psq, piece)  ((psq)[C_PIECES*((piece)-1)])
#define  SqFindOne(psq, piece)    SqFindFirst(psq,piece)
#define  SqFindSecond(psq, piece) ((psq)[C_PIECES*((piece)-1)+1])
#define  SqFindThird(psq, piece)  ((psq)[C_PIECES*((piece)-1)+2])

//Undefine and define some common constants from "eval.h"
//because the Nalimov code uses them too.
#undef INF
#undef P
#undef N
#undef B
#undef R
#undef Q
#undef K
#define CINTERFACE
//Locking will be done manually
#ifdef SMP
# define SMP_UNDEFINED
# undef SMP
#endif
#ifdef _MSC_VER
 #pragma message("MSC compatible compiler detected -- turning off warning 4716")
 #pragma warning( disable : 4127 )
#endif
#include "Nalimov/tbindex.cpp"
#ifdef SMP_UNDEFINED
# undef SMP_UNDEFINED
# define SMP
#endif
//reinclude everything
#ifdef INF
# undef INF
#endif
#define INF MATE
#ifdef P
# undef P
#endif
#define P 0
#ifdef N
# undef N
#endif
#define N 1
#ifdef B
# undef B
#endif
#define B 2
#ifdef R
# undef R
#endif
#define R 3
#ifdef Q
# undef Q
#endif
#define Q 4
#ifdef K
# undef K
#endif
#define K 5

#ifdef _MSC_VER
#include "bitinstructions.c"
#endif

extern int probeNalimov(const board* const pos, int* result)
{
	if(popcnt(pos->AllPieces)>NalimovNumPieces) return (0);
	{
		int    rgiCounters[10];
		square sqEnP;
		INDEX  ind;
		int    iTb;
		color  side;
		int    fInvert;
		int    tbValue;

		//to help compute locations of pieces
		square WhitePieces[C_PIECES*5+1], BlackPieces[C_PIECES*5+1];
		square *psqW, *psqB;

		//Initialize rgiCounters
		/* rgiCounters is a 10-element integer array. Its zeroth element is number
		 * of white pawns on the board, first - number of white knights, ...,
		 * ninth - number of black queens.
		 */
		{
			rgiCounters[0] = matsig_getP(pos->matsig[WHITE]);
			rgiCounters[1] = matsig_getN(pos->matsig[WHITE]);
			rgiCounters[2] = matsig_getB(pos->matsig[WHITE]);
			rgiCounters[3] = matsig_getR(pos->matsig[WHITE]);
			rgiCounters[4] = matsig_getQ(pos->matsig[WHITE]);
			rgiCounters[5] = matsig_getP(pos->matsig[BLACK]);
			rgiCounters[6] = matsig_getN(pos->matsig[BLACK]);
			rgiCounters[7] = matsig_getB(pos->matsig[BLACK]);
			rgiCounters[8] = matsig_getR(pos->matsig[BLACK]);
			rgiCounters[9] = matsig_getQ(pos->matsig[BLACK]);
		}

		/**
		 *IDescFindFromCounters (int *)
		 *It takes in rgiCounters. It returns zero if no such TB registered,
		 *TB #, if TB found, or negative TB # if you asked for TB with 'wrong'
		 *colors, e.g. KPKRP.
		 */
		iTb = IDescFindFromCounters(rgiCounters);

		if(iTb == 0) return (0); //No TB found 
		if(iTb > 0)  //TB found
		{
			side = pos->side;
			fInvert = false;
			psqW = WhitePieces;
			psqB = BlackPieces;
		}
		else //Found TB with 'wrong' colors
		{
			side = pos->xside;
			fInvert = true;
			psqW = BlackPieces;
			psqB = WhitePieces;
			iTb = -iTb;
		}

		//Make sure the tablebase requested is registered during initialization
		if(!FRegistered(iTb,side)) return (0);

		//square of en-passant capture, or XX if there's no en-passant available
		sqEnP = pos->EP?pos->EP:XX;


		//initialize WhitePieces and BlackPieces
		SqFindKing(WhitePieces) = pos->KingPos[WHITE];
		SqFindKing(BlackPieces) = pos->KingPos[BLACK];
		{
			U64 temp = piecesWHITE(*pos,P);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(WhitePieces,x_piecePawn));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(WhitePieces,x_piecePawn));
					if(temp)
						GetBitAndClear(temp,SqFindThird(WhitePieces,x_piecePawn));
				}
			}
			temp = piecesWHITE(*pos,N);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(WhitePieces,x_pieceKnight));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(WhitePieces,x_pieceKnight));
					if(temp)
						GetBitAndClear(temp,SqFindThird(WhitePieces,x_pieceKnight));
				}
			}
			temp = piecesWHITE(*pos,B);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(WhitePieces,x_pieceBishop));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(WhitePieces,x_pieceBishop));
					if(temp)
						GetBitAndClear(temp,SqFindThird(WhitePieces,x_pieceBishop));
				}
			}
			temp = piecesWHITE(*pos,R);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(WhitePieces,x_pieceRook));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(WhitePieces,x_pieceRook));
					if(temp)
						GetBitAndClear(temp,SqFindThird(WhitePieces,x_pieceRook));
				}
			}
			temp = piecesWHITE(*pos,Q);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(WhitePieces,x_pieceQueen));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(WhitePieces,x_pieceQueen));
					if(temp)
						GetBitAndClear(temp,SqFindThird(WhitePieces,x_pieceQueen));
				}
			}
			//Now do black
			temp = piecesBLACK(*pos,P);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(BlackPieces,x_piecePawn));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(BlackPieces,x_piecePawn));
					if(temp)
						GetBitAndClear(temp,SqFindThird(BlackPieces,x_piecePawn));
				}
			}
			temp = piecesBLACK(*pos,N);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(BlackPieces,x_pieceKnight));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(BlackPieces,x_pieceKnight));
					if(temp)
						GetBitAndClear(temp,SqFindThird(BlackPieces,x_pieceKnight));
				}
			}
			temp = piecesBLACK(*pos,B);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(BlackPieces,x_pieceBishop));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(BlackPieces,x_pieceBishop));
					if(temp)
						GetBitAndClear(temp,SqFindThird(BlackPieces,x_pieceBishop));
				}
			}
			temp = piecesBLACK(*pos,R);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(BlackPieces,x_pieceRook));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(BlackPieces,x_pieceRook));
					if(temp)
						GetBitAndClear(temp,SqFindThird(BlackPieces,x_pieceRook));
				}
			}
			temp = piecesBLACK(*pos,Q);
			if(temp)
			{
				GetBitAndClear(temp,SqFindFirst(BlackPieces,x_pieceQueen));
				if(temp)
				{
					GetBitAndClear(temp,SqFindSecond(BlackPieces,x_pieceQueen));
					if(temp)
						GetBitAndClear(temp,SqFindThird(BlackPieces,x_pieceQueen));
				}
			}
		}

		/**
		 * PfnIndCalc(iTb, side) is a macro that returns the function that
		 * calculates the index of the position in the TB. It is then
		 * subsequently used to index the tablebase.
		 */
		ind = PfnIndCalc(iTb,side) (psqW,psqB,sqEnP,fInvert);

		/**
		 * L_TbtProbeTable(int iTb, color side, INDEX index) probes the TB and returns
		 * either a result or 'L_bev_broken' if some error happened. If the specified TB
		 * was loaded to memory, the function will return result immediately. Otherwise,
		 * it will try to find needed value in TB cache; in the worst case it have to read
		 * a chunk of a file from the disk.
		 */
		tbValue = L_TbtProbeTable(iTb,side,ind);
		if(tbValue == L_bev_broken) return (0);

		//Fix the score
		if(tbValue>0) *result = NALIMOV_MATE-tbValue;
		else if (tbValue<0) *result = -(NALIMOV_MATE+tbValue);
		else *result = 0;
		return (1);
	}
}

extern int initNalimov(char* path, unsigned long cache_size)
{
	if(!strcmp(path,"")) return (0);
	NalimovNumPieces = IInitializeTb(path);
	if(NalimovCache!=NULL){ free(NalimovCache); NalimovCache=NULL; }
	if(NalimovNumPieces)
	{
		NalimovCache = (void*)malloc(cache_size);
		FTbSetCacheSize(NalimovCache, cache_size);
		return (1);
	}
	return (0);
}

extern void finalizeNalimov()
{
	VTbCloseFiles();
	if(NalimovCache!=NULL){ free(NalimovCache); NalimovCache=NULL; }
}