/**
*Buzz Chess Engine
*timemanager.c
*
*Copyright (C) 2008 Pradu Kannan.
**/

#if (defined(_WIN32) || defined(_WIN64))
	#define VC_EXTRALEAN
	#define WIN32_LEAN_AND_MEAN
	#define NOMINMAX
	#include <windows.h>
#endif

#ifdef _POSIX
	#include <sys/time.h>
#endif

#include "timemanager.h"
#include "xboard.h"
#include "log.h"


unsigned int st=0;           /*seconds*/
unsigned int mps=0;          /*moves per TC*/
unsigned int nextTC=0;       /*milliseconds*/
unsigned int inc=0;          /*seconds*/
unsigned int myTime=300000;  /*milliseconds*/
unsigned int opponentTime=5; /*milliseconds*/
time_ms startTime=0;         /*milliseconds*/
time_ms endTime=0;           /*milliseconds*/
time_ms searchTime=0;		 /*time search used,needed,if we switch from ponder to "normal"*/

//Private Data
bool infTime=false;

time_ms getms()
{
	#if (defined(_WIN32) || defined(_WIN64))
		return GetTickCount();
	#endif
	#ifdef _POSIX
		struct timeval tv;
		gettimeofday(&tv,NULL);
		return (((time_ms) tv.tv_sec)*1000) + (time_ms) tv.tv_usec/1000;
	#endif
}

void infiniteTime()
{
	infTime=true;
	printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - Infinite Time\n");
}

/*
void finiteTime()
{
	infTime=false;
	printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - Finite Time\n");
}
*/

void finiteTime()
{
   if (!sd)
   {
       infTime=false;
       printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - Finite Time\n");
   }
   else
   {
       printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - Error trying to set finite time on sd\n");
   }
}
void setStartTime()
{
	startTime=getms();
}

void setEndTime()
{
	time_ms localmyTime;
	time_ms timeToUse;
	localmyTime=myTime;

	//if(localmyTime>((time_ms)BUFFER_TIME*10)) localmyTime-=BUFFER_TIME;

	if(sd)
	{
		infTime=true;
		return;
	}

	if(st) timeToUse=st*1000;
	else
	{
		if(mps)
		{
			unsigned int movesLeft=mps-((moveNumber/2)+(compSide==WHITE && moveNumber%2))%mps+BUFFER_MPS;
			timeToUse=localmyTime/movesLeft;
		}
		else{
			timeToUse=localmyTime/TIME_DIVIDER;
			// a pondermiss? extend searchtime by 2,its crude but seems to work in selfplay :P
			if(ponder && ponderHashKey==C64(0x1234567890ABCDEF))//not a ponder search
				timeToUse= localmyTime/(TIME_DIVIDER/2);
		}
	}
	if(inc)
		timeToUse += (inc*1000-BUFFER_INC);
	//adjust my time
	if(!timeToUse) timeToUse=1000;
	
	//startTime is set when we receive 'time xxx' from winboard
	printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - time to use %u (from myTime %u) ms\n",timeToUse,localmyTime);
	endTime=startTime+timeToUse;
	printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - set end time to %u\n",endTime);
#if 0

	if(!st)
	{
		//Increment
		localmyTime+=inc*1000;
		if(mps)
		{
			unsigned int moves_played;
			if(compSide==WHITE)
				moves_played = (moveNumber+1)/2;
			else
				moves_played = moveNumber/2;

			if(compSide==gamePosition[moveNumber].side 
				&& ((moves_played + 1)%mps)==0)
			{
				localmyTime+=nextTC;
				printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - Added %u ms for the next time control\n",nextTC);
			}
		}
		//Time used up
		if(((unsigned)timeToUse)<=localmyTime) localmyTime-=timeToUse;
		else localmyTime=0;
	}
	printLogHTML(HTMLLOG_COLOR_GREEN,"Timemanager - estimated myTime after play %u\n",localmyTime);

#endif
}

bool timeUp()
{
	if(infTime) return false;
	return (bool) ( getms()>= endTime);
}

void startsearchTimer(){
	searchTime=getms();
}

time_ms searchtimeUsed()
{
	return getms()-searchTime;
}
